local WEAPON_NONE = -1

local function safeGetMT(mt)
	local success, value = pcall(function() return type(mt) == "string" and _G[mt] or mt end)
	return success and value or nil
end

local function HL_CreateItem(mt, stats)
	local mobj = type(mt) == "string" and _G[mt] or mt
	if not mobj return end
	HL_PickupStats[mobj] = stats
end
local function HL_DefineWeapon(name, stats)
	HL_WpnStats[name] = stats
end

local function scaleFaccioloHealth(facc)
	local h = facc.health or 0

	if facc.phase == 1 then
		-- Phase 1 (short)
		return h * 50 -- Makes each point ~.357 power
	elseif facc.phase == 2 then
		-- Phase 2 (scalable)
		-- Optional: Add scaling based on actual player count
		local playercount = 0
		for player in players.iterate do
			if player.mo and player.mo.valid then
				playercount = $ + 1
			end
		end

		local base = 400
		local perplayer = 100
		local expectedmax = 146

		-- Normalize and scale
		local ratio = FRACUNIT * h / expectedmax
		local scaledhp = base + (perplayer * (playercount - 1))
		return FixedMul(scaledhp * FRACUNIT, ratio) / FRACUNIT
	end

	-- Fallback if we don't recognize the health
	return h * 10
end

local function CheckAddons()
	if DoomGuy
		-- DOOM:
		HL_CreateItem(safeGetMT(MT_ITEM_STIMPACK), {health = {give = 10}})
		HL_CreateItem(safeGetMT(MT_ITEM_HEALTHPACK), {health = {give = 25}})
		HL_CreateItem(safeGetMT(MT_ITEM_COMBAT_ARMOR), {armor = {set = "limit", maxmult = FRACUNIT*2, novox = true}})
		HL_CreateItem(safeGetMT(MT_ITEM_SECURITY_ARMOR), {armor = {set = "limit", maxmult = FRACUNIT, novox = true}})
		HL_CreateItem(safeGetMT(MT_ITEM_SOUL), {health = {give = "maxhp", maxmult = FRACUNIT*2}})
		HL_CreateItem(safeGetMT(MT_ITEM_MEGA), {health = {give = "limit", maxmult = FRACUNIT*2}, armor = {give = "limit", maxmult = FRACUNIT*2, novox = true}})
		HL_CreateItem(safeGetMT(MT_ITEM_INVULNERABILITY), {invuln = {set = 20*TICRATE}})
		HL_CreateItem(safeGetMT(MT_POWERUP_BERSERK), {berserk = INT32_MAX})
		HL_CreateItem(safeGetMT(MT_POWERUP_BACKPACK), {ammo = {type = {"bull","shel","rckt","cell"}, give = {10,4,1,20}}, doubleammo = true})
		HL_CreateItem(safeGetMT(MT_ITEM_HEALTH), {health = {give = 1, maxmult = FRACUNIT*2}})
		HL_CreateItem(safeGetMT(MT_ITEM_ARMOR), {armor = {give = 1, maxmult = FRACUNIT*2}})
		HL_CreateItem(safeGetMT(MT_AMMO_CLIP), {ammo = {type = "bull", give = 10}})
		HL_CreateItem(safeGetMT(MT_AMMO_CLIP_BOX), {ammo = {type = "bull", give = 50}})
		HL_CreateItem(safeGetMT(MT_AMMO_SHELL), {ammo = {type = "shel", give = 4}})
		HL_CreateItem(safeGetMT(MT_AMMO_SHELL_BOX), {ammo = {type = "shel", give = 20}})
		HL_CreateItem(safeGetMT(MT_AMMO_ROCKET), {ammo = {type = "rckt", give = 1}})
		HL_CreateItem(safeGetMT(MT_AMMO_ROCKET_BOX), {ammo = {type = "rckt", give = 5}})
		HL_CreateItem(safeGetMT(MT_AMMO_CELL), {ammo = {type = "cell", give = 20}})
		HL_CreateItem(safeGetMT(MT_AMMO_CELL_PACK), {ammo = {type = "cell", give = 200}})
		HL_CreateItem(safeGetMT(MT_WEAPON_CHAINSAW), {weapon = "doomchainsaw"})
		HL_CreateItem(safeGetMT(MT_WEAPON_PISTOL), {weapon = "doompistol"})
		HL_CreateItem(safeGetMT(MT_WEAPON_SHOTGUN), {weapon = "doomshotgun"})
		HL_CreateItem(safeGetMT(MT_WEAPON_SUPERSHOTGUN), {weapon = "doomsupershotgun"})
		HL_CreateItem(safeGetMT(MT_WEAPON_CHAINGUN), {weapon = "doomchaingun"})
		HL_CreateItem(safeGetMT(MT_WEAPON_ROCKETLAUNCHER), {weapon = "doomrpg"})
		HL_CreateItem(safeGetMT(MT_WEAPON_PLASMARIFLE), {weapon = "doomplasmarifle"})
		HL_CreateItem(safeGetMT(MT_WEAPON_BFG9000), {weapon = "doombfg9000"})

		HL_DefineWeapon("doomchainsaw", {
			viewmodel = "WP2",
			weaponclass = "doom",
			neverdenyuse = true,
			weaponslot = 1,
			priority = 999,
			doomwepswitch = true,
			primary = {
				ammo = "melee",
				clipsize = WEAPON_NONE,
				israycaster = true,
				shotcost = 0,
				damagemin = 2,
				damagemax = 20,
				damageincs = 2,
				horizspread = 0,
				vertspread = 0,
				firesound = sfx_sawful,
				firedelay = 4,
			},
			pickupsound = sfx_wpnup,
			switchsound = sfx_sawup,
			hitsound = sfx_sawhit,
			autoreload = true,
			maxdistance = 8,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Chainsaw (DOOM)",
		})
		HL_DefineWeapon("doompistol", {
			viewmodel = "WP2",
			weaponclass = "doom",
			weaponslot = 2,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 10,
				ammo = "bull", -- Holy shit, DoomGuy uses live bulls as ammunition?! (no
				refireusesspread = true,
				israycaster = true,
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 5,
				damagemax = 15,
				damageincs = 5,
				horizspread = 11*FRACUNIT/2,
				vertspread = 11*FRACUNIT/2,
				firesound = sfx_pist,
				firedelay = 14,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Pistol (DOOM)",
		})
		HL_DefineWeapon("doomshotgun", {
			viewmodel = "WP3",
			weaponclass = "doom",
			selectgraphic = "HL1HUDSHOTGUN",
			weaponslot = 3,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 4,
				israycaster = true,
				ammo = "shel",
				pellets = 7,
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 5,
				damagemax = 15,
				damageincs = 5,
				horizspread = 11*FRACUNIT/2,
				vertspread = 11*FRACUNIT/2,
				firesound = sfx_sht,
				firedelay = 41,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Shotgun (DOOM)",
		})
		HL_DefineWeapon("doomsupershotgun", {
			viewmodel = "WP3A",
			weaponclass = "doom",
			selectgraphic = "HL1HUDSHOTGUN",
			weaponslot = 3,
			priority = 998,
			doomwepswitch = true,
			primary = {
				pickupgift = 4,
				israycaster = true,
				ammo = "shel",
				pellets = 20,
				clipsize = WEAPON_NONE,
				shotcost = 2,
				damagemin = 5,
				damagemax = 15,
				damageincs = 5,
				horizspread = 11*FRACUNIT/2,
				vertspread = 11*FRACUNIT/2,
				firesound = sfx_ssg,
				firedelay = 48,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Super Shotgun",
		})
		HL_DefineWeapon("doomchaingun", {
			viewmodel = "WP4",
			weaponclass = "doom",
			selectgraphic = "HL1HUDMP5",
			weaponslot = 4,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 10,
				ammo = "bull",
				refireusesspread = true,
				israycaster = true,
				volley = 2,
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 5,
				damagemax = 15,
				damageincs = 5,
				horizspread = 11*FRACUNIT/2,
				vertspread = 11*FRACUNIT/2,
				firesound = sfx_pist,
				firedelay = 4,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Chaingun",
		})
		HL_DefineWeapon("doomrpg", {
			viewmodel = "WP3",
			weaponclass = "doom",
			selectgraphic = "HL1HUDRPG",
			weaponslot = 5,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 2,
				ammo = "rckt",
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 20,
				damagemax = 160,
				damageincs = 20,
				horizspread = 0,
				vertspread = 0,
				firesound = sfx_rklaun,
				firedelay = 18,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
			},
			realname = "Rocket Launcher (DOOM)",
		})
		HL_DefineWeapon("doomplasmarifle", {
			viewmodel = "WP3",
			weaponclass = "doom",
			selectgraphic = "HL1HUDGAUSS",
			weaponslot = 6,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 100,
				ammo = "cell",
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 5,
				damagemax = 40,
				damageincs = 5,
				horizspread = 0,
				vertspread = 0,
				firesound = sfx_plasma,
				firedelay = 3,
			},
			pickupsound = sfx_wpnup,
			globalfiredelay = {
				ready = 23,
				firepost = 20,
			},
			realname = "Plasma Rifle",
		})
		HL_DefineWeapon("doombfg9000", { -- UNFINISHED!! i mean most of these things are but this is majorly unfinished
			viewmodel = "WP3",
			weaponclass = "doom",
			selectgraphic = "HL1HUDTAU",
			weaponslot = 7,
			priority = 999,
			doomwepswitch = true,
			primary = {
				pickupgift = 100,
				firefunc = function(player, mystats)
					S_StartSound(player.mo, sfx_bfg)
					return true
				end,
				ammo = "cell",
				clipsize = WEAPON_NONE,
				shotcost = 1,
				damagemin = 100,
				damagemax = 800,
				damageincs = 100,
				horizspread = 0,
				vertspread = 0,
				firedelay = 10,
			},
			pickupsound = sfx_wpnup,
			autoreload = true,
			vmdlflip = false,
			globalfiredelay = {
				ready = 23,
				tilshootmobj = 30,
			},
			realname = "BFG9000",
		})
		HL_AmmoStats.rckt.shootmobj = MT_DROCKET
		HL_AmmoStats.cell.shootmobj = MT_DPLASMA
		HL.DoDoomguyAccomodations = true -- Because the Rocket Launcher uses variables that WILL break DoomGuy if I just... *spawn* the object
	end

	if duke_roboinfo then
		kombihl1viewmodels.DUKESHOT.animations.primaryfire.frameSounds = {nil, nil, nil, nil, nil, sfx_shtcck}
		local mightyboot = {
			ammo = "none",
			clipsize = -1,
			shotcost = 0,
			damage = 6,
			horizspread = 5*FRACUNIT,
			vertspread = 5*FRACUNIT,
			firedelay = 6,
		}
		HL_DefineWeapon("dukepistol", {
			viewmodel = "DUKEPIST",
			weaponclass = "duke3d",
			selectgraphic = "HL1HUD9MM",
			weaponslot = 1,
			priority = 999,
			primary = {
				ammo = "9mm",
				israycaster = true,
				pickupgift = 48,
				clipsize = 12,
				shotcost = 1,
				damage = 6,
				refireusesspread = true,
				horizspread = 5*FRACUNIT,
				vertspread = 5*FRACUNIT,
				firesound = sfx_pistol,
				reloadsound = sfx_pstcck,
				firedelay = 6,
			},
			secondary = mightyboot,
			pickupsound = sfx_pstcck,
			globalfiredelay = {
				ready = 9,
				reload = 30,
			},
			realname = "Pistol (Duke3D)",
		})
		HL_DefineWeapon("dukeshotgun", {
			viewmodel = "DUKESHOT",
			weaponclass = "duke3d",
			selectgraphic = "HL1HUDSHOTGUN",
			weaponslot = 2,
			priority = 999,
			primary = {
				ammo = "buckshot",
				israycaster = true,
				pickupgift = 48,
				clipsize = -1,
				shotcost = 1,
				damage = 10,
				pellets = 7,
				horizspread = 7*FRACUNIT,
				vertspread = 7*FRACUNIT,
				firesound = sfx_shtgun,
				firedelay = 36,
			},
			secondary = mightyboot,
			pickupsound = sfx_shtcck,
			globalfiredelay = {
				ready = 9,
				reload = 30,
			},
			realname = "Pistol (Duke3D)",
		})
		HL_DefineWeapon("dukechaingun", {
			viewmodel = "DUKESHOT",
			weaponclass = "duke3d",
			selectgraphic = "HL1HUDMP5",
			weaponslot = 3,
			priority = 999,
			primary = {
				ammo = "9mm",
				israycaster = true,
				pickupgift = 48,
				clipsize = -1,
				shotcost = 1,
				damage = 6,
				horizspread = 5*FRACUNIT,
				vertspread = 5*FRACUNIT,
				kickback = 5*FRACUNIT/2,
				firesound = sfx_dchain,
				firedelay = 4,
			},
			secondary = mightyboot,
			pickupsound = sfx_pstcck,
			globalfiredelay = {
				ready = 9,
				reload = 30,
			},
			realname = "Chaingun (Duke3D)",
		})
	end

	if sfx_wpist then
		HL_DefineWeapon("wolfknife", {
			viewmodel = "WOLFKNIF",
			weaponclass = "wolf3d",
			selectgraphic = "HL1HUD9MM",
			weaponslot = 1,
			priority = 999,
			primary = {
				ammo = "melee",
				israycaster = true,
				clipsize = -1,
				shotcost = 0,
				damagemin = 1,
				damagemax = 15,
				damageincs = 1,
				firesound = nil,
				firedelay = 12,
				maxdistance = 12,
				fireoffset = 6,
				firesound = sfx_wknif,
			},
			pickupsound = sfx_wammo,
			globalfiredelay = {
				ready = 9,
			},
			realname = "Knife (Wolf3D)",
		})
		HL_DefineWeapon("wolfpistol", {
			viewmodel = "WOLFPIST",
			weaponclass = "wolf3d",
			selectgraphic = "HL1HUD9MM",
			weaponslot = 2,
			priority = 999,
			primary = {
				ammo = "9mm",
				pickupgift = 6,
				israycaster = true,
				clipsize = -1,
				shotcost = 1,
				firesound = nil,
				firedelay = 12,
				fireoffset = 3,
				firesound = sfx_wpist,
				firehitfunc = function(player, victim)
					local dist2 = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y)

					if dist2 < (64*2*FRACUNIT)^2 then
						return P_RandomByte() >> 2
					elseif dist2 < (64*4*FRACUNIT)^2 then
						return P_RandomByte() / 6
					else
						local misschance = P_RandomByte() / 12
						-- Compare misschance to "distance in tiles", rounded down
						local approx_dist_tiles = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y) / (64*FRACUNIT)
						if misschance < approx_dist_tiles then
							return 0  -- Missed
						end
						return P_RandomByte() / 6
					end
				end
			},
			pickupsound = sfx_wammo,
			globalfiredelay = {
				ready = 9,
			},
			realname = "Pistol (Wolf3D)",
		})
		HL_DefineWeapon("wolfmachinegun", {
			viewmodel = "WOLFMACH",
			weaponclass = "wolf3d",
			selectgraphic = "HL1HUDMP5",
			weaponslot = 3,
			priority = 999,
			primary = {
				ammo = "9mm",
				pickupgift = 6,
				israycaster = true,
				clipsize = -1,
				shotcost = 1,
				firesound = nil,
				refiredelay = 2,
				firedelay = 8,
				fireoffset = 2,
				refireoffset = 0,
				firesound = sfx_wmgun,
				firehitfunc = function(player, victim)
					local dist2 = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y)

					if dist2 < (64*2*FRACUNIT)^2 then
						return P_RandomByte() >> 2
					elseif dist2 < (64*4*FRACUNIT)^2 then
						return P_RandomByte() / 6
					else
						local misschance = P_RandomByte() / 12
						-- Compare misschance to "distance in tiles", rounded down
						local approx_dist_tiles = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y) / (64*FRACUNIT)
						if misschance < approx_dist_tiles then
							return 0  -- Missed
						end
						return P_RandomByte() / 6
					end
				end
			},
			pickupsound = sfx_wmgpic,
			globalfiredelay = {
				ready = 9,
			},
			realname = "Machine Gun (Wolf3D)",
		})
		HL_DefineWeapon("wolfchaingun", {
			viewmodel = "WOLFCHAI",
			weaponclass = "wolf3d",
			selectgraphic = "HL1HUDMP5",
			weaponslot = 4,
			priority = 999,
			primary = {
				ammo = "9mm",
				pickupgift = 6,
				israycaster = true,
				clipsize = -1,
				shotcost = 1,
				firesound = nil,
				refiredelay = 0,
				firedelay = 8,
				fireoffset = 2,
				refireoffset = 0,
				volleyfireoffset = 2,
				volleyfiredelay = 2,
				volley = 2,
				firesound = sfx_wcgun,
				firehitfunc = function(player, victim)
					local dist2 = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y)

					if dist2 < (64*2*FRACUNIT)^2 then
						return P_RandomByte() >> 2
					elseif dist2 < (64*4*FRACUNIT)^2 then
						return P_RandomByte() / 6
					else
						local misschance = P_RandomByte() / 12
						-- Compare misschance to "distance in tiles", rounded down
						local approx_dist_tiles = R_PointToDist2(player.mo.x, player.mo.y, victim.x, victim.y) / (64*FRACUNIT)
						if misschance < approx_dist_tiles then
							return 0  -- Missed
						end
						return P_RandomByte() / 6
					end
				end
			},
			pickupsound = sfx_wcgpic,
			globalfiredelay = {
				ready = 9,
			},
			realname = "Chaingun (Wolf3D)",
		})
	end

	-- SONIC DOOM II:
	HL_SetMTStats(safeGetMT("MT_SD2_BUZZBOMBER"), {health = 400, dmgdmgmult = 2*FRACUNIT, flinches = true}, {min = 10, max = 80, increments = 10})
	HL_SetMTStats(safeGetMT("MT_SD2_COCONUTS"), {health = 60, dmgmult = 2*FRACUNIT, flinches = true}, {min = 8, max = 24, increments = 8})
	HL_SetMTStats(safeGetMT("MT_SD2_GROUNDER_PISTOL"), {health = 20, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 15, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_GROUNDER_SHOTGUN"), {health = 30, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 15, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_GROUNDER_CHAINGUN"), {health = 30, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 15, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_METALSONIC"), {health = 700, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 15, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_PSEUDOKNUCKLES"), {health = 300, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 15, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_PSEUDOFLICKY"), {health = 300, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 24, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_PSEUDOTAILS"), {health = 400, dmgmult = 2*FRACUNIT, flinches = true}, {min = 3, max = 24, increments = 3})
	HL_SetMTStats(safeGetMT("MT_SD2_VILE_FIRE"), 0, {dmg = 90})
	HL_SetMTStats(safeGetMT("MT_SD2_OVASHORT"), {health = 150, dmgmult = 2*FRACUNIT, flinches = true}, {min = 4, max = 40, increments = 4})
	HL_SetMTStats(safeGetMT("MT_SD2_OVASHORT_SHADOW"), {health = 150, dmgmult = 2*FRACUNIT, flinches = true}, {min = 4, max = 40, increments = 4})
	HL_SetMTStats(safeGetMT("MT_SD2_OVASHOT"), 0, {min = 8, max = 64, increments = 8})
	HL_SetMTStats(safeGetMT("MT_SD2_OVARED"), {health = 1000, dmgmult = 2*FRACUNIT, flinches = true}, {min = 10, max = 80, increments = 10})
	HL_SetMTStats(safeGetMT("MT_SD2_OVAGRAY"), {health = 1000, dmgmult = 2*FRACUNIT, flinches = true}, {min = 10, max = 80, increments = 10})
	HL_SetMTStats(safeGetMT("MT_SD2_PSEUDOSUPER"), {health = 500, dmgmult = 2*FRACUNIT, flinches = true}, 0)
	HL_SetMTStats(safeGetMT("MT_SD2_PSEUDOSUPER_BALL"), 0, {min = 5, max = 40, increments = 5})
	HL_SetMTStats(safeGetMT("MT_SD2_REDMETALSONIC"), {health = 4000, dmgmult = 2*FRACUNIT, flinches = true}, {min = 10, max = 80, increments = 10})
	HL_SetMTStats(safeGetMT("MT_SD2_ROCKET"), 0, {min = 20, max = 160, increments = 20})

	if Silverhorn then
		HL_SetMTStats(MT_UNA_FORCE, {health = 300})

		addHook("ThinkFrame", function()
			local mo
			local facc

			for mobj in mobjs.iterate() do
				if mobj.type == MT_FACCIOLO_BOSS and mobj.facc then
					mo = mobj
					facc = mobj.facc
					break
				end
			end

			if not (mo and facc) then return end

			if (facc.health or 0) > (facc.hloldhealth or 0) then
				mo.hl.health = scaleFaccioloHealth(facc)
				mo.hl.maxhealth = mo.hl.health
				print("Refresh...", facc.health .. " health makes " .. mo.hl.health .. "(Old HP was " .. facc.hloldhealth .. ")")
			end
			facc.hloldhealth = facc.health
		end)
	end
	
	if sfx_BALPOF and sfx_puyo1 and sfx_pspawn and sfx_blkout then
		HL_SetMTStats(MT_SF94_OLDNPC, {health = INT32_MAX})
		HL_SetMTStats(MT_SF94_OLDSTANDNPC, {health = INT32_MAX})
	end
	
	if sfx_scream and sfx_tdsee and sfx_redscr then
		HL.DoTDForestAccomodations = true
	end

	if MM then
		MM.addHook("MovementSpeedCap", function(player)
			if player.mo.skin != "kombifreeman" then return end
			return 60*FRACUNIT
		end)
		MM.addHook("CorpseThink", function(corpse)
			if corpse.skin != "kombifreeman" then return end
			local player = players[corpse.playerid]
			local pcam = player.awayviewmobj
			if pcam and pcam.valid then
				P_MoveOrigin(corpse, player.awayviewmobj.x,player.awayviewmobj.y,corpse.z)
				corpse.backup = {x = pcam.momx, y = pcam.momy, z = pcam.momz}
			elseif corpse.backup then
				corpse.momx, corpse.momy, corpse.momz = corpse.backup.x, corpse.backup.y, corpse.backup.z
			end
		end)
	end
end

addHook("AddonLoaded", function()
	if OLDC and OLDC.SkinFullNames and not OLDC.SkinFullNames["kombifreeman"]
		if P_RandomChance(FRACUNIT/4)
			OLDC.SkinFullNames["kombifreeman"] = "JOHN HALFLIFE"
		elseif P_RandomChance(FRACUNIT/4)
			OLDC.SkinFullNames["kombifreeman"] = "GORDON FREEMAN THE THEORETICAL PHYSICIST"
		elseif P_RandomChance(FRACUNIT/4)
			OLDC.SkinFullNames["kombifreeman"] = "GORDON THE FREEMAN"
		else
			OLDC.SkinFullNames["kombifreeman"] = "GORDON FREEMAN"
		end
	end
	CheckAddons()
end)

CheckAddons()