local function SafeFreeSlot(...)
	for _,slot in ipairs({...})
		if not rawget(_G, slot) freeslot(slot) end
	end
end

local function warn(player, str)
	CONS_Printf(player, "\130WARNING: \128" .. tostring(str));
end

-- I Hate those Koopalings killed self with worldspawn

local skin = "kombifreeman"
local FLASH_DRAIN_TIME = FRACUNIT*6/5
local FLASH_CHARGE_TIME = FRACUNIT/5

SafeFreeSlot("SPR2_CRCH", "S_PLAY_FREEMCROUCH", "S_PLAY_FREEMCROUCHMOVE",
"MT_HLRAY",
"sfx_hldeny","sfx_hlflas","sfx_hlspra",
"sfx_pwepst","sfx_pwepsl","sfx_pwepcl","sfx_pwepen",
"sfx_hlfal1","sfx_hlfal2","sfx_hlfal3",
"sfx_frpai1","sfx_frpai2","sfx_frpai3","sfx_frpai4","sfx_frpai5",
"sfx_hlla1","sfx_hlla2","sfx_hlla3","sfx_hlla4",
"sfx_hlmed1","sfx_hlmed2")
sfxinfo[sfx_hldeny].caption = "\135Can't Use\x80" -- for SOME reason using the usual hex for this caption turns it cyan and eats the first two proper letters
sfxinfo[sfx_hlmed1].caption = "MEDIC!!"
sfxinfo[sfx_hlmed2].caption = "MEDIC!!"

states[S_PLAY_FREEMCROUCHMOVE] = {
	sprite = SPR_PLAY,
	frame = SPR2_CRCH|FF_ANIMATE,
	tics = -1,
	var1 = 3,
	var2 = 12,
	nextstate = S_PLAY_FREEMCROUCH
}

states[S_PLAY_FREEMCROUCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_CRCH,
	tics = -1,
	var1 = 3,
	var2 = 4,
	nextstate = S_PLAY_FREEMCROUCH
}

spr2defaults[SPR2_CRCH] = SPR2_WALK

local function SpawnSpray(mo)
	local spr  = P_SpawnPlayerMissile(mo, MT_SPRAY)
	spr.pangle = mo.angle
	spr.tracer = mo
	spr.z      = $+(mo.player.viewheight/2)
	return spr
end

-- Constants
local CATEGORY_COUNT = 10

-- Returns an empty weapon selection structure
local function emptySelection()
	local ws = {}
	for i = 0, 9 do ws[i] = 0 end
	return {
		weapons = {},
		weaponcount = 0,
		wepslotamounts = ws
	}
end

-- Returns the index of the first usable weapon, or nil if none
local function getFirstUsableIndex(sel)
	if not sel or not sel.weapons then return nil end
	for i, w in ipairs(sel.weapons) do
		if w.usable then return i end
	end
	return nil
end

-- Steps through the current bucket without wrapping
local function cycleWithinCategory(sel, startIndex, step)
	local cnt = sel and sel.weaponcount or 0
	if cnt < 2 then return nil end

	local idx = startIndex + step
	while idx >= 1 and idx <= cnt do
		if sel.weapons[idx].usable then return idx end
		idx = idx + step
	end

	return nil
end

-- Steps through the current bucket with wraparound
local function cycleWithinCategoryWrap(sel, startIndex, step)
	local cnt = sel and sel.weaponcount or 0
	if cnt < 1 then return nil end

	local idx = startIndex
	for _ = 1, cnt do
		idx = ((idx - 1 + step) % cnt) + 1
		if sel.weapons[idx].usable then return idx end
	end

	return nil
end

-- Tries to switch to a new weapon category with at least one usable weapon
local function switchCategory(currentCat, dir, player)
	for i = 1, CATEGORY_COUNT do
		local offset = (dir == 1 and i or -i)
		local newCat = (currentCat + offset + CATEGORY_COUNT) % CATEGORY_COUNT
		local sel = HL_GetWeapons(HL_WpnStats, newCat, player) or emptySelection()
		if getFirstUsableIndex(sel) then
			return newCat, sel
		end
	end
	local fallback = HL_GetWeapons(HL_WpnStats, 11, player)
	return CATEGORY_COUNT + 1, fallback or emptySelection() -- fallback if nothing usable
end

-- Main weapon cycling function (non-wrapping within category)
local function HL_CycleWeapon(player, direction)
	if not player or not player.hlinv.weapons then return end
	player.kombiaccessinghl1menu = true

	local dirstep = direction == "next" and 1 or -1
	local cat = player.kombihl1category or 0
	local sel = player.selectionlist or HL_GetWeapons(HL_WpnStats, cat, player) or emptySelection()
	local first = getFirstUsableIndex(sel)

	if sel.weaponcount == 0 or not first then
		local nc, ns = switchCategory(cat, dirstep, player)
		player.kombiprevhl1category = cat
		player.kombihl1category = nc
		player.selectionlist = ns
		player.kombihl1wpn = getFirstUsableIndex(ns) or 0
		return
	end

	local cur = player.kombihl1wpn or first
	local newIndex = cycleWithinCategory(sel, cur, dirstep)

	if newIndex then
		player.kombihl1wpn = newIndex
		return
	end

	local nc, ns = switchCategory(cat, dirstep, player)
	player.kombiprevhl1category = cat
	player.kombihl1category = nc
	player.selectionlist = ns

	if direction == "next" then
		player.kombihl1wpn = getFirstUsableIndex(ns) or 0
	else
		for i = #ns.weapons, 1, -1 do
			if ns.weapons[i].usable then
				player.kombihl1wpn = i
				break
			end
		end
		player.kombihl1wpn = player.kombihl1wpn or 0
	end
end

COM_AddCommand("invnext", function(player)
	if gamestate ~= GS_LEVEL then
		CONS_Printf(player, "Can't do that right now.")
		return
	end
	HL_CycleWeapon(player, "prev")
end)

COM_AddCommand("invprev", function(player)
	if gamestate ~= GS_LEVEL then
		CONS_Printf(player, "Can't do that right now.")
		return
	end
	HL_CycleWeapon(player, "next")
end)

/*
violence_ablood 1	enable blood (0 will improve perfomance some, but you won't see any blood)
violence_agibs 1	enable gibs (0 will improve performance some, but you won't see body chunks)
violence_hblood 1	enable more blood (0 will improve perfomance some, but you won't see as much blood)
violence_hgibs 1	enable more gibs (0 will improve performance some, but you won't see as many body chunks)
*/
local function printTable(data, prefix)
	prefix = prefix or ""
	if type(data) == "table"
		for k, v in pairs(data or {}) do
			local key = prefix .. k
			if type(v) == "table" then
				CONS_Printf(server, "key " .. key .. " = a table:")
				printTable(v, key .. ".")
			else
				CONS_Printf(server, "key " .. key .. " = " .. tostring(v))
			end
		end
	else
		CONS_Printf(server, data)
	end
end

-- Define a helper that takes the base command name, and two functions: one for the '+' variant and one for the '-' variant.
local function RegisterDualCommand(cmd, onPress, onRelease)
	-- Create the press command ("+<cmd>")
	COM_AddCommand("+" .. cmd, function(player, ...)
		if not player.mo then
			return
		end
		-- Register that the player is holding the command
		player.hlcmds = player.hlcmds or {}
		player.hlcmds[cmd] = true
		-- Execute the onPress behavior if provided
		if onPress then onPress(player, ...) end
	end)

	-- Create the release command ("-<cmd>")
	COM_AddCommand("-" .. cmd, function(player, ...)
		if not player.mo then
			return
		end
		-- Mark the command as released
		player.hlcmds = player.hlcmds or {}
		player.hlcmds[cmd] = false
		-- Execute the onRelease behavior if provided
		if onRelease then onRelease(player, ...) end
	end)
end

local MAX_USE_ANGLE = ANGLE_45 -- how wide of a horizontal angle we can search
local MAX_USE_DIST = USERANGE -- How far the check can go before it's too far
RegisterDualCommand("use",
	function(player)
		S_StartSound(player.mo, sfx_hldeny)
	end)

RegisterDualCommand("duck")
RegisterDualCommand("jump")
RegisterDualCommand("reload")
RegisterDualCommand("attack")
RegisterDualCommand("attack2")
RegisterDualCommand("speed")
RegisterDualCommand("lookup",
	function(player) -- onPress
		player.hl.pitchLook = ($ or 0) + 1
	end,
	function(player) -- onRelease
		player.hl.pitchLook = ($ or 0) - 1
	end)

RegisterDualCommand("lookdown",
	function(player) -- onPress
		player.hl.pitchLook = ($ or 0) - 1
	end,
	function(player) -- onRelease
		player.hl.pitchLook = ($ or 0) + 1
	end)
RegisterDualCommand("left",
	function(player) -- onPress
		player.hl.yawLook = ($ or 0) + 1
	end,
	function(player) -- onRelease
		player.hl.yawLook = ($ or 0) - 1
	end)

RegisterDualCommand("right",
	function(player) -- onPress
		player.hl.yawLook = ($ or 0) - 1
	end,
	function(player) -- onRelease
		player.hl.yawLook = ($ or 0) + 1
	end)


function string.trim(s)
	return s:match("^%s*(.-)%s*$")
end

local function stripOuterQuotes(s)
    if type(s) ~= "string" then return s end
    while s:sub(1,1)=='"' and s:sub(-1)=='"' do
        s = s:sub(2,-2)
    end
    return s
end

-- Helper function: Get what the command is supposed to be ("+cmd" for dual, "cmd" for normal)
local function getBinding(bind)
    if type(bind) == "string" then
        bind = stripOuterQuotes(bind)
        local cmds = {}

        for piece in bind:gmatch("[^;]+") do
            piece = piece:trim()
            local dual = false
            if piece:sub(1,1)=="+" then
                dual = true
                piece = piece:sub(2)
            end
            table.insert(cmds, {command=piece, dual=dual})
        end
        return cmds
    end
    return bind
end

-- Reports if the player is able to do anything (quick-fixes some custom things)
local function playerHasControl(player)
return not (
  player.exiting
  or gamestate ~= GS_LEVEL
  or player.powers[pw_nocontrol]
  or P_PlayerInPain(player)
  or (player.pflags & PF_STASIS)
  or (player.pflags & PF_FULLSTASIS)
  or (player.powers[pw_carry] > CR_NONE)
  or (player.playerstate ~= PST_LIVE)
) end

-- True character printout
local keyalias = {
	KEY123 = "{",
	KEY124 = "|",
	KEY125 = "}",
	KEY126 = "~",
	TILDE = "`",
	space = " ",
}

-- Used to automatically switch to shiftkeys if we're holding SHIFT
local shiftdown = {
	l = false,
	r = false
}

local function isShiftDown()
	return shiftdown.l or shiftdown.r
end

-- Aliased key binds.
-- Note: By prefixing the command with "+", you automatically mark the command as dual. The associated command should be prefixed with a "-"!!
-- e.g. "+something" "-something"
local defaultKeyBinds = {
	q				 = "lastinv",		-- SHIELD
	e				 = "+use",			-- TOSSFLAG
	r				 = "+reload",		-- CUSTOM 1
	lctrl			 = "+duck",			-- SPIN
	lshift			 = "+speed",		-- CUSTOM 3
	f				 = "impulse 100",	-- CUSTOM 2
	t				 = "impulse 201",
	z				 = "saveme",
	space			 = "+jump",
	mouse1			 = "+attack",
	mouse2			 = "+attack2",
	pgdn			 = "+lookdown",
	pgup			 = "+lookup",
	left			 = "+left",
	right			 = "+right",
	["0"]			 = "slot0",
	["1"]			 = "slot1",			-- WPN SLOT 1
	["2"]			 = "slot2",			-- WPN SLOT 2
	["3"]			 = "slot3",			-- WPN SLOT 3
	["4"]			 = "slot4",			-- WPN SLOT 4
	["5"]			 = "slot5",			-- WPN SLOT 5
	["6"]			 = "slot6",			-- WPN SLOT 6
	["7"]			 = "slot7",			-- WPN SLOT 7
	["8"]			 = "slot8",
	["9"]			 = "slot9",
	["wheel 1 up"]	 = "invprev",
	["wheel 1 down"] = "invnext",
}

-- Custom command that basically lets us have certain commands on any button
COM_AddCommand("hl_bind", function(player, key, command)
	if not command then
		CONS_Printf(player, "Usage: hl_bind <key> <command>", "Example: bind e +use")
		return
	end
	if not player.keyBinds then return end
	player.keyBinds[key] = command
	saveTableToFile(HL.KEYBINDS_PATH, player.keyBinds)
	CONS_Printf(player, "Bound key '" .. key .. "' to command '" .. command .. "'")
end)

COM_AddCommand("hl_unbind", function(player, key)
	if not key then
		CONS_Printf(player, "Usage: hl_unbind <key>", "Example: unbind e")
		return
	end
	if not player.keyBinds then return end
	player.keyBinds[key] = nil
	saveTableToFile(HL.KEYBINDS_PATH, player.keyBinds)
end)

-- KeyDown hook function
local function OnKeyDown(keyevent)
	-- print(keyevent.name)
	if not consoleplayer then return end
	local isshift = isShiftDown()
	if keyevent.name == "TILDE" and not isshift then return end
	if keyevent.name == "lshift" then shiftdown.l = true end
	if keyevent.name == "rshift" then shiftdown.r = true end
	local bindEntry = consoleplayer.keyBinds and consoleplayer.keyBinds[keyevent.name]
	local isTyping = consoleplayer.hl and consoleplayer.hl.typingmode
	local typeMode = consoleplayer.hl and consoleplayer.hl.typingmode == 1 and "say" or "sayteam"
	if isTyping then
		local totype = tostring(keyalias[keyevent.name] or keyevent.name)
		local typed = isshift and input.keyNumToName(input.shiftKeyNum(keyevent.num)) or totype
		if isshift then typed = keyalias[$] or $ end
		typed = tostring(typed)
		if typed == "enter" then
			COM_BufInsertText(consoleplayer, typeMode .. ' "' .. consoleplayer.hl.typing .. '"')
			COM_BufInsertText(consoleplayer, "messagemode")
		elseif typed == "backspace" then
			consoleplayer.hl.typing = string.sub($, 1, -2)
		elseif input.keyNumPrintable(keyevent.num)
			consoleplayer.hl.typing = ($ or "") .. typed
		end
		return true
	end
	if not consoleplayer.realmo then return end
	if (consoleplayer.playerstate ~= PST_LIVE) and bindEntry then
		local bindings = getBinding(bindEntry)
		for _, binding in ipairs(bindings) do
			if not keyevent.repeated then
				if binding.command != "jump" then return end
				if binding.dual then
					COM_BufInsertText(consoleplayer, "+" .. binding.command)
				else
					COM_BufInsertText(consoleplayer, binding.command)
				end
			end
		end
	end
	if not playerHasControl(consoleplayer) then return end
	if consoleplayer.realmo.skin != skin then return end
	if bindEntry then
		local bindings = getBinding(bindEntry)
		for _, binding in ipairs(bindings) do
			if not keyevent.repeated then
				if binding.dual then
					COM_BufInsertText(consoleplayer, "+" .. binding.command)
				else
					COM_BufInsertText(consoleplayer, binding.command)
				end
			end
		end
		return true
	end
	return false	-- No binding: allow normal processing.
end

local function isButtonBind(keyevent)
	for i = 0, NUM_GAMECONTROLS - 1 do
		if input.keyNumToName(input.gameControlToKeyNum(i)) == keyevent.name then
			return true
		end
	end
	return false
end

-- KeyUp hook function, only needed for dual commands.
local function OnKeyUp(keyevent)
	if not consoleplayer then return end
	if not consoleplayer.realmo then return end
	if not playerHasControl(consoleplayer) then return end
	if consoleplayer.realmo.skin != skin then return end
	if keyevent.name == "lshift" then shiftdown.l = false end
	if keyevent.name == "rshift" then shiftdown.r = false end
	local bindEntry = consoleplayer.keyBinds and consoleplayer.keyBinds[keyevent.name]
	if bindEntry then
		local bindings = getBinding(bindEntry)
		for _, binding in ipairs(bindings) do
			if binding.dual then
				COM_BufInsertText(consoleplayer, "-" .. binding.command)
			end
		end
		return not isButtonBind(keyevent)
	end
	return false
end

-- Register the hooks.
addHook("KeyDown", OnKeyDown)
addHook("KeyUp", OnKeyUp)
/*
local PLAYER_FATAL_FALL_SPEED = FixedMul(2769450, FRACUNIT*8/6)
local PLAYER_MAX_SAFE_FALL_SPEED = FixedMul(1107000, FRACUNIT*8/6)
local DAMAGE_FOR_FALL_SPEED = FixedDiv(100*FRACUNIT,(PLAYER_FATAL_FALL_SPEED - PLAYER_MAX_SAFE_FALL_SPEED))
local PLAYER_FALL_PUNCH_THRESHOLD = FixedMul(668000, FRACUNIT*8/6)

#define PLAYER_FATAL_FALL_SPEED		1024// approx 60 feet
#define PLAYER_MAX_SAFE_FALL_SPEED	580// approx 20 feet
#define DAMAGE_FOR_FALL_SPEED		(float) 100 / ( PLAYER_FATAL_FALL_SPEED - PLAYER_MAX_SAFE_FALL_SPEED )// damage per unit per second.
#define PLAYER_MIN_BOUNCE_SPEED		200
#define PLAYER_FALL_PUNCH_THRESHHOLD (float)350 // won't punch player's screen/make scrape noise unless player falling at least this fast.
*/
local function HLToDoom(hlUnit)
    return FixedMul(hlUnit, HL.ConvRatio)
end

local PLAYER_FATAL_FALL_SPEED = 45*FRACUNIT
local PLAYER_MAX_SAFE_FALL_SPEED = 26*FRACUNIT
local DAMAGE_FOR_FALL_SPEED = FixedDiv(100*FRACUNIT,(PLAYER_FATAL_FALL_SPEED - PLAYER_MAX_SAFE_FALL_SPEED))
local PLAYER_FALL_PUNCH_THRESHOLD = 18*FRACUNIT

local nofalldmgmaps = {
	["12 convoy assault"] = function(player)
		return player.awayviewtics
	end
}

local function HL_GetFallDamage(fallSpeed, player)
	local mapkey = gamemap .. " " .. string.lower(G_BuildMapTitle(gamemap) or "")
	local nofalldmg = nofalldmgmaps[mapkey]

	-- Skip fall damage if map says so, if in water, or player flags it off
	if (nofalldmg and (type(nofalldmg) ~= "function" or nofalldmg(player)))
	or (player.mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER) ~= 0)
	or (player.hl and (player.hl.nofalldmg or player.hl.fallcount))
	then
		if player.hl.fallcount then player.hl.fallcount = $-1 end
		return {dmg = 0, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
	end

	local fallspeed_abs = abs(fallSpeed)

	-- Respect cv_kombifalldamage.value before checking for safe/fatal speed thresholds
	local falldmgmode = cv_kombifalldamage.value
	if falldmgmode == -1 then -- "Dont"
		return {dmg = 0, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
	elseif falldmgmode == 0 then -- "Fixed"
		return {dmg = fallspeed_abs >= PLAYER_MAX_SAFE_FALL_SPEED and 10 or 0, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
	end

	-- Handle safe fall (no damage) and fatal fall (max damage)
	if fallspeed_abs <= PLAYER_MAX_SAFE_FALL_SPEED then
		return {dmg = 0, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
	elseif fallspeed_abs >= PLAYER_FATAL_FALL_SPEED then
		-- If not "On", still apply damage based on custom setting.
		if falldmgmode == 1 then -- "On"
			local calcDamage = FixedMul(fallspeed_abs - PLAYER_MAX_SAFE_FALL_SPEED, DAMAGE_FOR_FALL_SPEED)
			return {dmg = min(FixedInt(calcDamage), 100), fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
		else
			return {dmg = 100, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
		end
	end

	-- Handle fall damage between safe and fatal fall
	if falldmgmode == 1 then -- "On"
		local calcDamage = FixedMul(fallspeed_abs - PLAYER_MAX_SAFE_FALL_SPEED, DAMAGE_FOR_FALL_SPEED)
		return {dmg = min(FixedInt(calcDamage), 100), fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
	end

	-- Default fixed value
	return {dmg = 10, fallpunch = (fallSpeed >= PLAYER_FALL_PUNCH_THRESHOLD)}
end

addHook("PlayerThink", function(player)
	if not player.hl then return end
	if not player.mo then return end
	if player.mo.skin != skin then return end

	if player.hl.pitchLook then
		player.aiming = $ + FixedAngle((cv_pitchspeed.value / 60) * player.hl.pitchLook)
	end

	if player.hl.yawLook then
		player.mo.angle = $ + FixedAngle((cv_yawspeed.value / 60) * player.hl.yawLook)
	end

	if not player.mo then return end

	if (player.pflags & PF_NOCLIP) then
		player.mo.flags = $ | MF_NOGRAVITY | MF_NOCLIPHEIGHT | MF_NOCLIP -- physics in QuakePhysics
	elseif player.hl.noclippedtic then
		player.mo.flags = $ & ~(MF_NOGRAVITY | MF_NOCLIPHEIGHT | MF_NOCLIP)
		player.awayviewaiming = 0
		player.awayviewtics = 0
		player.hl.noclippedtic = nil
	end

	-- Handle impulse 201 call (since SRB2 seems to whine when we do it in the command itself)
	if player.hl and player.hl.spraying then
		SpawnSpray(player.mo)
		player.hl.spraying = false
	end

	-- If Freeman, handle fall damage
	if player.realmo.skin == "kombifreeman" then
		local mapkey = gamemap .. " " .. string.lower(G_BuildMapTitle(gamemap) or "")
		local nofalldmg = nofalldmgmaps[mapkey]
		if player.hl and (nofalldmg and (type(nofalldmg) ~= "function" or nofalldmg(player)))
			player.hl.fallcount = 1
		end
		if player.mo.eflags & MFE_JUSTHITFLOOR then
			local fallhurt = HL_GetFallDamage(abs(player.kombifallz or 0), player)
			if fallhurt.dmg then
				P_DamageMobj(player.mo, nil, nil, fallhurt.dmg, HL.DMG.FALL)
				S_StartSound(player.mo,P_RandomRange(sfx_hlfal1,sfx_hlfal3))
				if player.mo.hl.health <= 0 then
					player.mo.killfallvel = player.kombifallz
				end
			end
			if fallhurt.fallpunch then
				player.hl = $ or {}
				player.hl.punchangle.z = (ANG350+ANG2+ANG2+ANG2) >> 16
			end
			player.kombifallz = 0
		elseif not P_IsObjectOnGround(player.mo) then
			-- Record the fall velocity for future fall damage checks
			player.kombifallz = player.mo.momz
		end
	end
end)

local laddertextures = {
    CUT10    = true,
	ConL     = true,
	HLLADDER = true,
}

local function intervalsIntersect(aBottom, aTop, bBottom, bTop)
    -- True if they share *any* non‑zero overlap
    return (aTop  > bBottom)
       and (aBottom < bTop)
end

addHook("MobjMoveBlocked", function(mo, thing, line)
    -- Basic sanity
    if not (mo and mo.valid and mo.player and line) then
        if mo.player then mo.player.hl = mo.player.hl or {} end
        return
    end

    local hl = mo.player.hl or {}
    mo.player.hl = hl

    -- Determine which side of the line we're on
    local sideIndex = P_PointOnLineSide(mo.x, mo.y, line)
    local mysec     = (sideIndex == 0) and line.frontsector or line.backsector
    local oppsec    = (sideIndex == 0) and line.backsector  or line.frontsector
    local sidedef   = (sideIndex == 0) and line.frontside   or line.backside

    local pBot = mo.z
    local pTop = mo.z + mo.height

    -- Prepare texture lookup table
    local ladderTexs = laddertextures

    -- Fast-path: iterate both sectors' ffloors lists in one go
    local sectors = { mysec, oppsec }
    for si = 1, 2 do
        local sec = sectors[si]
        if sec then
            for ff in sec.ffloors() do
                if ff.valid and (ff.flags & FOF_EXISTS) ~= 0 then
                    local fBot = ff.bottomheight
                    local fTop = ff.topheight

                    if intervalsIntersect(fBot, fTop, pBot, pTop) then
                        local ms = ff.master.frontside
                        if ms then
                            -- grab all three texture names once
                            local mid  = R_TextureNameForNum(ms.midtexture)
                            local top  = R_TextureNameForNum(ms.toptexture)
                            local bot  = R_TextureNameForNum(ms.bottomtexture)

                            if ladderTexs[mid] or ladderTexs[top] or ladderTexs[bot] then
                                hl.climb             = true
                                hl.climbing          = line
                                hl.climbing_is_front = sideIndex
                                hl.ladderbottom     = fBot
                                hl.laddertop        = fTop
                                return
                            end
                        end
                    end
                end
            end
        end
    end

    -- No FOF ladder found — fall back on sidedef textures
    if sidedef then
        local mt = R_TextureNameForNum(sidedef.midtexture)
        local tt = R_TextureNameForNum(sidedef.toptexture)
        local bt = R_TextureNameForNum(sidedef.bottomtexture)

        if ladderTexs[mt] or ladderTexs[tt] or ladderTexs[bt] then
            hl.climb             = true
            hl.climbing          = line
            hl.climbing_is_front = sideIndex

            if mysec ~= oppsec then
                -- two different sectors
                hl.ladderbottom = mysec.floorheight
                hl.laddertop    = oppsec.floorheight
            else
                -- single-sided line
                hl.ladderbottom = mysec.floorheight
                hl.laddertop    = mysec.ceilingheight
            end

            return
        end
    end

    -- Not a ladder: clear stored data
    hl.climb        = nil
    hl.climbing     = nil
    hl.ladderbottom = nil
    hl.laddertop    = nil
end, MT_PLAYER)

local function ClosestPointOnSegment_t(px, py, x1, y1, x2, y2)
    local dx, dy = x2 - x1, y2 - y1
    local denom = FixedMul(dx,dx) + FixedMul(dy,dy)
    if denom == 0 then
        return x1, y1, 0, true
    end

    -- projection numerator
    local num = FixedMul(px - x1, dx) + FixedMul(py - y1, dy)
    -- unclamped t in fixed [0..FRACUNIT] space
    local tUnc = FixedDiv(num, denom)

    local outside = false
    local t = tUnc
    if tUnc < 0 then
        t, outside = 0, true
    elseif tUnc > FRACUNIT then
        t, outside = FRACUNIT, true
    end

    local cx = x1 + FixedMul(t, dx)
    local cy = y1 + FixedMul(t, dy)
    return cx, cy, t, outside
end

addHook("PreThinkFrame", function()
    for player in players.iterate do
        local mo = player.mo
        if not (mo and mo.valid) then
            if mo then mo.flags = mo.flags & ~MF_NOGRAVITY end
            continue
        end

        local cmd = player.cmd
        local hl  = player.hl or {}
        player.hl = hl

        if hl.climb and hl.climbing then
            -- vertical check
            local bottom = min(hl.ladderbottom, hl.laddertop)
            local top    = max(hl.ladderbottom, hl.laddertop)
            if mo.z < bottom or mo.z > top then
                hl.climb = false
				hl.sndtick = 0
                mo.flags = mo.flags & ~MF_NOGRAVITY
                cmd.forwardmove, cmd.sidemove = 0,0
                continue
            end

            -- corner-segment detach
            local x1,y1 = hl.climbing.v1.x, hl.climbing.v1.y
            local x2,y2 = hl.climbing.v2.x, hl.climbing.v2.y
            local ladangle = R_PointToAngle2(x1,y1, x2,y2)

            -- ladder-plane normal
            local nx,ny = cos(ladangle+ANGLE_90), sin(ladangle+ANGLE_90)
            if not hl.climbing_is_front then nx,ny = -nx,-ny end

            -- ladder tangent (for horizontal)
            local tx,ty = cos(ladangle), sin(ladangle)

            local radius = mo.radius
            local thr   = FixedMul(radius, abs(nx)+abs(ny))
            local SAFE  = FRACUNIT/8

            local anyOn = false
            for _,off in ipairs({{radius,radius},{radius,-radius},{-radius,radius},{-radius,-radius}}) do
                local px,py = mo.x+off[1], mo.y+off[2]
                local cx,cy,t,out = ClosestPointOnSegment_t(px,py, x1,y1, x2,y2)
                if not out then
                    local dx,dy = px-cx, py-cy
                    local dist  = FixedSqrt(FixedMul(dx,dx)+FixedMul(dy,dy))
                    if dist <= thr+SAFE then
                        anyOn = true
                        break
                    end
                end
            end
            if not anyOn then
                hl.climb = false
				hl.sndtick = 0
                mo.flags = mo.flags & ~MF_NOGRAVITY
                cmd.forwardmove, cmd.sidemove = 0,0
                continue
            end

            -- climb movement + sound
            mo.flags = mo.flags | MF_NOGRAVITY

            -- scale inputs into fixed speeds
            local ZSPEED = 6*FRACUNIT
            local XSPEED = 6*FRACUNIT
            local f = cmd.forwardmove * FRACUNIT
            local s = cmd.sidemove    * FRACUNIT

            -- sound timer (16-tic interval)
            hl.sndtick = hl.sndtick or 0
            if f != 0 or s != 0 then
                hl.sndtick = hl.sndtick + 1
                if hl.sndtick >= 16 then
                    hl.sndtick = 0
                    local sfx = sfx_hlla1 + P_RandomRange(0,3)
                    S_StartSound(mo, sfx)
                end
            end

            -- map into velocities
            local fv = FixedMul(f, FixedDiv(ZSPEED, 50*FRACUNIT))
            -- band-aid invert sidemove so right>0 moves you right
            local rv = FixedMul(-s, FixedDiv(XSPEED, 50*FRACUNIT))

            cmd.forwardmove, cmd.sidemove = 0,0

            -- jump off
            if cmd.buttons & BT_JUMP ~= 0 then
				local ladderSpeed = HLToDoom(200*FRACUNIT)
                mo.momx = FixedMul(nx, ladderSpeed)
                mo.momy = FixedMul(ny, ladderSpeed)
				P_SetObjectMomZ(mo, 4*FRACUNIT)
                hl.climb = false
				hl.sndtick = 0
                mo.flags = mo.flags & ~MF_NOGRAVITY
                continue
            end

            -- build world-space vel exactly like HL
            local a = mo.angle
            local vpnx,vpny = cos(a), sin(a)
            local vrx, vry  = cos(a+ANGLE_90), sin(a+ANGLE_90)
            local vx = FixedMul(vpnx,fv) + FixedMul(vrx,rv)
            local vy = FixedMul(vpny,fv) + FixedMul(vry,rv)

            -- decompose: lateral XY + −normal→Z
            local normal = FixedMul(vx,nx) + FixedMul(vy,ny)
            local lx = vx - FixedMul(nx, normal)
            local ly = vy - FixedMul(ny, normal)

            mo.momx = lx
            mo.momy = ly
            mo.momz = -normal
        else
            mo.flags = mo.flags & ~MF_NOGRAVITY
        end
    end
end)

-- Constants for battery drain/recharge rates
local DRAIN_RATE = FRACUNIT / 42
local RECHARGE_RATE = FRACUNIT / 7
local MAX_BATTERY = 100 * FRACUNIT

addHook("PlayerSpawn",function(player)
	if not player.mo return end
	if secondarydisplayplayer == player
	and splitscreen
	and player.mo.skin == "kombifreeman"
		camera2.chase = false
	elseif player.mo.skin == "kombifreeman"
		camera.chase = false
	end
	player.hlinv = $ or {}
	player.hl = $ or {}
	player.hlcmds = {}
	player.voxBuffer = {{}, {}}
	player.hl.punchangle = $ or {x = 0, y = 0, z = 0}
	player.hl.suitvoicewait = {}
	player.hl.dmgicons = {}
	player.hl.cooking = false
	player.keyBinds = loadTableFromFile(HL.KEYBINDS_PATH, defaultKeyBinds)
	for k,v in pairs(player.keyBinds) do
		player.keyBinds[k] = stripOuterQuotes(v)
	end
	saveTableToFile(HL.KEYBINDS_PATH, player.keyBinds)
	player.hl.spray = loadTableFromFile(HL.SPRAY_CONFIG_PATH, {spray = "lambda", color = nil})
	if not cv_deathmatch.value then
		player.hlinv.ammo = $ or {
			["9mm"] = 68, -- the only value we have set in stone... (def 68)
			["357"] = 36,
			buckshot = 125,
			bolt = 50,
			grenade = 10,
			satchel = 5,
			tripmine = 5,
			melee = -1, -- these two require values so that any user error with default ammo types won't be pinned on me
			none = -1, -- ^ because yeah of COURSE it'd throw an error at you if you tried to decrement it IT WASN'T EVEN SUPPOSED TO BE DECREMENTED
		}
		player.hlinv.wepclips = $ or {}
		player.hlinv.weapons = $ or {
			crowbar = true,
			["9mmhandgun"] = true,
			["357"] = true,
			["mp5"] = true,
			shotgun = true,
			crossbow = true,
			rpg = true,
			handgrenade = true,
			satchel = true,
		}
	else
		player.hlinv.ammo = {
			["9mm"] = 68,
			melee = -1,
			none = -1,
		}
		player.hlinv.wepclips = {}
		player.hlinv.weapons = {
			crowbar = true,
			["9mmhandgun"] = true,
			["357"] = true,
		}
	end
	player.hlinv.hevsuit = true
	player.pickuphistory = {index = 1}
	player.hl1deadtimer = 0
	player.hl.curwep = startweapon or $ or "crowbar"
	player.hl.flashlightbattery = MAX_BATTERY
	if not player.hlinv.wepclips[player.hl.curwep] then
		local clipsize = HL_WpnStats[player.hl.curwep].primary and HL_WpnStats[player.hl.curwep].primary.clipsize or -1
		local clipsize2 = HL_WpnStats[player.hl.curwep].secondary and HL_WpnStats[player.hl.curwep].secondary.clipsize or -1
		player.hlinv.wepclips[player.hl.curwep] = {primary = clipsize, secondary = clipsize2}
	end
	HL_ChangeViewmodelState(player, "ready", "idle")
	if cv_ljmspawn.value then
		ApplyPickupStats(player, {longjump = true})
	end
end)

addHook("PlayerCanEnterSpinGaps", function(player)
	if not player.mo then return end
	local shouldCrouch = (player.cmd.buttons & BT_SPIN) or (player.hlcmds and player.hlcmds.duck) or not (P_CheckPosition(player.mo, player.mo.x, player.mo.y, true) and (player.prevpos and P_CheckPosition(player.mo, player.prevpos.x, player.prevpos.y, true)))
	if shouldCrouch then return true end
end)

addHook("PlayerHeight", function(player)
	if player.mo.skin != "kombifreeman" then return end
	return player.hl.height
end)

local srb2defviewheight = 89 * FRACUNIT -- def == 41
local CROUCH_TRANS_TIME = TICRATE * 3 / 10

local function HL_IsPhysDisabled(player)
	return player.hl.nophys or player.hl.nofalldmg or player.hl.nograv
end

addHook("PlayerThink", function(player)
    if not player.mo then return end

	-- Automatically set spriteyscale because we don't have sprites for crouching yet
	if player.mo.skin == "kombifreeman" then
		player.mo.spriteyscale = FixedDiv(player.mo.height, 60*FRACUNIT)
		player.hl = $ or {}
		player.hl.stretched = true
	elseif player.hl.stretched then
		player.hl.stretched = false
		player.mo.spriteyscale = FRACUNIT
	end

    if player.mo.skin ~= skin then return end
    player.hl = player.hl or {}

	if player.hl.zoomed then
		player.fovadd = $ + 90 * FRACUNIT
	end

    -- get target heights
    local normalH = P_GetPlayerHeight(player)
    local crouchH = P_GetPlayerSpinHeight(player)
    local targetCrouch = 
        (player.cmd.buttons & BT_SPIN) ~= 0 
        or (player.hlcmds and player.hlcmds.duck)
        or not (
            P_CheckPosition(player.mo, player.mo.x, player.mo.y, true)
            and (player.prevpos and P_CheckPosition(player.mo, player.prevpos.x, player.prevpos.y, true))
        )

    -- init timer if starting a new crouch/stand
    if targetCrouch and not player.hl.wasCrouching then
        player.hl.crouchTimer = 0
    elseif not targetCrouch and player.hl.wasCrouching then
        player.hl.crouchTimer = CROUCH_TRANS_TIME
    end

    -- advance or rewind the timer
    if targetCrouch and player.hl.crouchTimer < CROUCH_TRANS_TIME then
        player.hl.crouchTimer = player.hl.crouchTimer + 1
    elseif not targetCrouch and player.hl.crouchTimer > 0 then
		if player.hl.crouchTimer > CROUCH_TRANS_TIME / 2 then
			player.hl.crouchTimer = CROUCH_TRANS_TIME / 2
		end
        player.hl.crouchTimer = player.hl.crouchTimer - 1
    end

    -- compute interpolated height
    local h = ease.linear(
        FixedDiv(player.hl.crouchTimer, CROUCH_TRANS_TIME), 
        normalH, 
        crouchH
    )

	local delta = abs(normalH - crouchH)

	if player.hl.crouchTimer == 1 and targetCrouch then
		-- just began crouching
		if not P_IsObjectOnGround(player.mo) and (player.mo.eflags & MFE_JUSTHITFLOOR) == 0 then
			player.mo.z = $ + delta * P_MobjFlip(player.mo)
		end
	elseif player.hl.crouchTimer == (CROUCH_TRANS_TIME / 2) - 1 and not targetCrouch then
		-- just began uncrouching
		if not P_IsObjectOnGround(player.mo) and (player.mo.eflags & MFE_JUSTHITFLOOR) == 0 then
			player.mo.z = $ - delta * P_MobjFlip(player.mo)
		end
	end

    -- apply the new height & viewheight
    player.hl.height    = h
    player.viewheight   = player.hl.height - 8*FRACUNIT

    -- adjust speed
    player.normalspeed = skins[player.realmo.skin].normalspeed
    if player.hlcmds and player.hlcmds.speed then
        player.normalspeed = FixedMul($, FRACUNIT*5/8)
    end
    if targetCrouch then
        player.cmd.forwardmove = $ / 3
        player.cmd.sidemove = $ / 3
    end

    -- animations
    local moving = abs(player.mo.momx) > 0 or abs(player.mo.momy) > 0
	if player.hl.wasCrouching != targetCrouch then
		if player.hl.crouchTimer == CROUCH_TRANS_TIME then
			-- fully crouched
			if moving then
				player.realmo.state = S_PLAY_FREEMCROUCHMOVE
			else
				player.realmo.state = S_PLAY_FREEMCROUCH
			end
		elseif player.hl.crouchTimer == 0 then
			-- fully standing
			player.realmo.state = S_PLAY_STND
		end
	end

    -- remember for next tick
    player.hl.wasCrouching = targetCrouch
    player.prevpos = { x = player.mo.x, y = player.mo.y }

	-- Flashlight thinker
	if player.hl.flashlightbattery <= 0 then
		player.hl.flashlight = false
		S_StartSound(player.mo, sfx_hlflas)
	end

	if player.hl.flashlight then
		if player.hl.flashlightbeam and player.hl.flashlightbeam.valid then
			player.hl.flashlightbeam.flags2 = $&~MF2_DONTDRAW
		end
		player.hl.flashlightbattery = player.hl.flashlightbattery - DRAIN_RATE
		P_SpawnPlayerMissile(player.mo, MT_HLFLASHLIGHTBEAM)
		if player.hl.flashlightbattery < 0 then
			player.hl.flashlightbattery = 0
		end
	else
		if player.hl.flashlightbeam and player.hl.flashlightbeam.valid then
			player.hl.flashlightbeam.flags2 = $|MF2_DONTDRAW
		end
		if player.hl.flashlightbattery < MAX_BATTERY then
			player.hl.flashlightbattery = player.hl.flashlightbattery + RECHARGE_RATE
			if player.hl.flashlightbattery > MAX_BATTERY then
				player.hl.flashlightbattery = MAX_BATTERY
			end
		end
	end

	-- Disable falling damage under these conditions (maybe because of script breaks, or is unplayable otherwise)
	-- Functionally different from nofalldmgmaps, as this is checked for no matter what map we're on (so we don't have to add every map in existence)
	if player.powers[pw_carry] == CR_MACESPIN then
		player.hl = $ or {}
		player.hl.nophys = true
		player.hl.nofalldmg = true
		player.hl.nograv = true
	end

	if (player.mo.eflags & MFE_SPRUNG) then
		player.hl = $ or {}
		player.hl.nograv = true
	end

	-- ...But clear the associated variables when we don't need them anymore.
	if player.hl and HL_IsPhysDisabled(player) and P_IsObjectOnGround(player.mo) then
		player.hl = $ or {}
		player.hl.nophys = false
		player.hl.nofalldmg = false
		player.hl.nograv = false
	end
	
	if player.hl.spraydelay then player.hl.spraydelay = $ - 1 end
	if player.mo.sprite2 != SPR2_STND then return end
	player.mo.tics = 105
	player.mo.state = S_PLAY_STND
end)

addHook("PlayerThink", function(player)
	local history = player.pickuphistory or {}
	local killfeed = player.killfeed or {}
	local dmgicons = player.hl and player.hl.dmgicons or {}

	-- decrement timers and remove expired entries
	for id, info in pairs(history) do
		if id == "index" then continue end
		if type(info) != "table" or not info.time or info.time <= 0 then
			history[id] = nil
		else
			info.time = $ - 1
		end
	end

	for id, info in pairs(killfeed) do
		if type(info) != "table" or not info.time or info.time <= 0 then
			killfeed[id] = nil
		else
			info.time = $ - 1
		end
	end

	for id, info in pairs(dmgicons) do
		if type(info) != "table" or not info.time or info.time <= 0 then
			dmgicons[id] = nil
		else
			info.time = $ - 1
		end
	end
end)

local function printTable(data, prefix)
	prefix = prefix or ""
	if type(data) == "table"
		if data == {} then
			print("[Empty table]")
		else
			for k, v in pairs(data or {}) do
				local key = prefix .. k
				if type(v) == "table" then
					print("key " .. key .. " = a table:")
					printTable(v, key .. ".")
				else
					print("key " .. key .. " = " .. tostring(v))
				end
			end
		end
	else
		print(data)
	end
end

addHook("PlayerThink", function(player)
	if not player.mo return end
	if player.mo.skin != skin then return end

	player.hl.messages = $ or {}
	local messages = player.hl.messages

	for i, msg in ipairs(messages) do
		msg.clock = $ + 1
		if msg.clock > msg.disappear_time then
			messages[i] = nil
		end
	end

	-- Add Silverhorn title message after regaining control
	if mapheaderinfo[gamemap].mindscape
	and MindscapeSessionFlags
	and (MindscapeSessionFlags & MS_SEENINTRO)
	and not player.hl.silverhorn_title_shown
	then
		player.hl.silverhorn_title_shown = true
		table.insert(messages, convertToHLMessage(player, "SILVERHORNTITLE"))
	end
end)

local function hasUsableWeapons(sel)
	return getFirstUsableIndex(sel) ~= nil
end

addHook("PlayerThink", function(player)
    if not player.mo then return end
	if player.mo.skin != skin then return end

    -- Selection via BT_WEAPONNEXT/PREV
    if not player.kombipressingselkeys then
        if player.cmd.buttons & BT_WEAPONNEXT then
            HL_CycleWeapon(player, "prev")
            player.kombipressingselkeys = true
        elseif player.cmd.buttons & BT_WEAPONPREV then
            HL_CycleWeapon(player, "next")
            player.kombipressingselkeys = true
        end
    else
        player.kombipressingselkeys = false
    end

    -- Selection via slot keys
    local slotKey = player.desiredSlot
    local mask = player.cmd.buttons & BT_WEAPONMASK
    if mask ~= 0 then
        slotKey = mask
    end

    if slotKey ~= nil and not player.kombipressingwpnkeys then
        -- remember if this really is the “first open”
        local firstOpen = not player.kombiaccessinghl1menu

        local prevCat = player.kombihl1category or -1
        local newCat  = slotKey
        player.kombiprevhl1category = prevCat
        player.kombihl1category     = newCat

        -- get the new selection list
        local sel = HL_GetWeapons(HL_WpnStats, newCat, player) or emptySelection()
        player.selectionlist = sel

        if sel.weaponcount > 0 and hasUsableWeapons(sel) then
            if firstOpen then
                -- on first open always pick the first slot
                player.kombihl1wpn = 1
            else
                -- same-category wrap or new-category default
                if newCat ~= prevCat then
                    player.kombihl1wpn = getFirstUsableIndex(sel) or 0
                else
                    local cur = player.kombihl1wpn or 1
                    player.kombihl1wpn = cycleWithinCategoryWrap(sel, cur, 1) or cur
                end
            end
        else
            player.kombihl1wpn = 0
        end

        -- play sound, mark state
        player.kombiaccessinghl1menu = true
        S_StartSound(player.mo,
            (newCat ~= prevCat or firstOpen)
              and sfx_pwepst or sfx_pwepsl
        )
        player.kombipressingwpnkeys = true
        player.desiredSlot = nil
    elseif player.kombipressingwpnkeys then
        player.kombipressingwpnkeys = false
    end

	if player.powers[pw_shield] then
		if (player.mo.hl.armor < player.mo.hl.maxarmor) then
			local amount = 20
			if (player.powers[pw_shield] == SH_PINK) then
				amount = 25
			elseif (player.powers[pw_shield] == SH_PITY) then
				amount = 15
			elseif (player.powers[pw_shield] == SH_WHIRLWIND) then
				amount = 25
			elseif (player.powers[pw_shield] == SH_ARMAGEDDON) then
				amount = 30
			elseif (player.powers[pw_shield] == SH_ELEMENTAL) then
				amount = 18
			elseif (player.powers[pw_shield] == SH_ATTRACT) then
				amount = 10
			elseif (player.powers[pw_shield] == SH_FLAMEAURA) then
				amount = 35
			elseif (player.powers[pw_shield] == SH_BUBBLEWRAP) then
				amount = 28
			elseif (player.powers[pw_shield] == SH_THUNDERCOIN) then
				amount = 40
			end

			if (player.powers[pw_shield] & SH_FORCE) then
				amount = amount + (5 * ((player.powers[pw_shield] & SH_FORCEHP) + 1))
			end
			if (player.powers[pw_shield] & SH_FIREFLOWER) then
				amount = amount + 10
			end

			player.mo.hl.armor = min(player.mo.hl.armor + (amount * FRACUNIT), player.mo.hl.maxarmor)
		end
		player.powers[pw_shield] = 0
	end
end)

local function HL_GetDamage(inf)
	if not HL1_DMGStats[inf.type] then return end
	local dmgstats = HL1_DMGStats[inf.type]
	local objdamage = dmgstats and dmgstats.damage or {}

	if objdamage.min and objdamage.max then
		local max = objdamage.max
		local min = objdamage.min
		local increment = objdamage.increments
		local divisor = increment or min
		return (P_RandomByte() % (max / divisor) + 1) * divisor
	else
		return objdamage.dmg
	end
end

local damageTypeMaps = {
	[DMG_CRUSHED] = HL.DMG.CRUSH|HL.DMG.TIMEBASED,
	[DMG_ELECTRIC] = HL.DMG.SHOCK|HL.DMG.TIMEBASED,
	[DMG_NUKE] = HL.DMG.BLAST|HL.DMG.TIMEBASED,
	[DMG_FIRE] = HL.DMG.BURN|HL.DMG.TIMEBASED,
	[DMG_DROWNED] = HL.DMG.DROWN|HL.DMG.TIMEBASED,
	[DMG_SPACEDROWN] = HL.DMG.DROWN|HL.DMG.TIMEBASED,
	[DMG_WATER] = HL.DMG.DROWN|HL.DMG.TIMEBASED,
	[DMG_SPIKE] = HL.DMG.SLASH,
	[DMG_DEATHPIT] = HL.DMG.FALL|HL.DMG.TIMEBASED,
}

local SUIT_NEXT_IN_30SEC = function(line, player)
	player.hl.suitvoicewait[line] = 30*TICRATE
end

local SUIT_NEXT_IN_1MIN = function(line, player)
	player.hl.suitvoicewait[line] = 1*TICRATE*60
end

local SUIT_NEXT_IN_5MIN = function(line, player)
	player.hl.suitvoicewait[line] = 5*TICRATE*60
end

local SUIT_NEXT_IN_10MIN = function(line, player)
	player.hl.suitvoicewait[line] = 10*TICRATE*60
end

local SUIT_NEXT_IN_30MIN = function(line, player)
	player.hl.suitvoicewait[line] = 30*TICRATE*60
end

local function DoDamageAndBursts(target, inf, damage, damagetype)
	local prevHealth = target.hl.health
    local kill = HL_DamageGordon(target, inf, damage, damagetype)

    -- On kill, burst emeralds/weapons/flags
    if kill then
        P_PlayerEmeraldBurst(target.player, false)
        P_PlayerWeaponAmmoBurst(target.player)
        P_PlayerFlagBurst(target.player, false)
    else
        -- compute trivial/major/critical on *post*-damage health
        local curHealth = target.hl.health
        local isTrivial = curHealth > 75 or damage < 5
        local isMajor   = damage > 25
        local isCritical= curHealth < 30

        -- save damage bits for HUD
        target.player.hl.bitsDamageType = damagetype
        target.player.hl.bitsHUDDamage  = -1

        -- iterate through each damage type
        local bits = damagetype
        local didOne = true
        while bits ~= 0
          and (not isTrivial or (bits & HL.DMG.TIMEBASED) ~= 0)
          and didOne
        do
            didOne = false

            if (bits & HL.DMG.CLUB) ~= 0 then
                if isMajor then
                    FVox_WarnDamage("HEV_DMG4", target.player, SUIT_NEXT_IN_30SEC)  -- minor fracture
                end
                bits = bits & ~HL.DMG.CLUB
                didOne = true
            end

            if (bits & (HL.DMG.FALL | HL.DMG.CRUSH)) ~= 0 then
                if isMajor then
                    FVox_WarnDamage("HEV_DMG5", target.player, SUIT_NEXT_IN_30SEC)  -- major fracture
                else
                    FVox_WarnDamage("HEV_DMG4", target.player, SUIT_NEXT_IN_30SEC)  -- minor fracture
                end
                bits = bits & ~(HL.DMG.FALL | HL.DMG.CRUSH)
                didOne = true
            end

            if (bits & HL.DMG.BULLET) ~= 0 then
                if damage > 5 then
                    FVox_WarnDamage("HEV_DMG6", target.player, SUIT_NEXT_IN_30SEC)  -- blood loss detected
                end
                bits = bits & ~HL.DMG.BULLET
                didOne = true
            end

            if (bits & HL.DMG.SLASH) ~= 0 then
                if isMajor then
                    FVox_WarnDamage("HEV_DMG1", target.player, SUIT_NEXT_IN_30SEC)  -- major laceration
                else
                    FVox_WarnDamage("HEV_DMG0", target.player, SUIT_NEXT_IN_30SEC)  -- minor laceration
                end
                bits = bits & ~HL.DMG.SLASH
                didOne = true
            end

            if (bits & HL.DMG.SONIC) ~= 0 then
                if isMajor then
                    FVox_WarnDamage("HEV_DMG2", target.player, SUIT_NEXT_IN_1MIN)   -- internal bleeding
                end
                bits = bits & ~HL.DMG.SONIC
                didOne = true
            end

            if (bits & (HL.DMG.POISON | HL.DMG.PARALYZE)) ~= 0 then
                FVox_WarnDamage("HEV_DMG3", target.player, SUIT_NEXT_IN_1MIN)      -- blood toxins detected
                bits = bits & ~(HL.DMG.POISON | HL.DMG.PARALYZE)
                didOne = true
            end

            if (bits & HL.DMG.ACID) ~= 0 then
                FVox_WarnDamage("HEV_DET1", target.player, SUIT_NEXT_IN_1MIN)     -- hazardous chemicals detected
                bits = bits & ~HL.DMG.ACID
                didOne = true
            end

            if (bits & HL.DMG.NERVEGAS) ~= 0 then
                FVox_WarnDamage("HEV_DET0", target.player, SUIT_NEXT_IN_1MIN)     -- biohazard detected
                bits = bits & ~HL.DMG.NERVEGAS
                didOne = true
            end

            if (bits & HL.DMG.RADIATION) ~= 0 then
                FVox_WarnDamage("HEV_DET2", target.player, SUIT_NEXT_IN_1MIN)     -- radiation detected
                bits = bits & ~HL.DMG.RADIATION
                didOne = true
            end

            if (bits & HL.DMG.SHOCK) ~= 0 then
                FVox_WarnDamage("HEV_SHOCK", target.player, SUIT_NEXT_IN_30SEC)
                bits = bits & ~HL.DMG.SHOCK
                didOne = true
            end

            if (bits & HL.DMG.BURN) ~= 0 then
                FVox_WarnDamage("HEV_FIRE", target.player, SUIT_NEXT_IN_30SEC)
                bits = bits & ~HL.DMG.BURN
                didOne = true
            end
        end

        -- view punch
        target.player.hl.punchangle = target.player.hl.punchangle or {}
        target.player.hl.punchangle.x = -2

        -- first time we take major damage
        if not isTrivial and isMajor and prevHealth >= 75 then
            FVox_WarnDamage("HEV_MED1", target.player, SUIT_NEXT_IN_30MIN)   -- automedic on
            FVox_WarnDamage("HEV_HEAL7", target.player, SUIT_NEXT_IN_30MIN)  -- morphine shot
        end

        -- escalating into critical
        if not isTrivial and isCritical and prevHealth < 75 then
            if curHealth < 6 then
                FVox_WarnDamage("HEV_HLTH3", target.player, SUIT_NEXT_IN_10MIN)  -- near death
            elseif curHealth < 20 then
                FVox_WarnDamage("HEV_HLTH2", target.player, SUIT_NEXT_IN_10MIN)  -- health critical
            end

            if (P_RandomRange(0,3) == 0) and prevHealth < 50 then
                FVox_WarnDamage("HEV_DMG7", target.player, SUIT_NEXT_IN_5MIN)     -- seek medical attention
            end
        end

        -- time‑based damage warnings
        if (damagetype & HL.DMG.TIMEBASED) ~= 0 and prevHealth < 75 then
            if prevHealth < 50 then
                if P_RandomRange(0,3) == 0 then
                    FVox_WarnDamage("HEV_DMG7", target.player, SUIT_NEXT_IN_5MIN) -- seek medical attention
                end
            else
                FVox_WarnDamage("HEV_HLTH1", target.player, SUIT_NEXT_IN_10MIN)   -- health dropping
            end
        end
    end

    return kill
end

addHook("MobjDamage", function(target, inf, src, dmg, dmgType)
    if target.skin ~= "kombifreeman" then
        return
    end

    -- world damage (no inf, no src)
    if not inf and not src then
        local baseDamage = ((dmgType and not damageTypeMaps[dmgType]) and dmg) or 15
        local baseType     = damageTypeMaps[dmgType & ~DMG_CANHURTSELF] or damageTypeMaps[dmgType] or dmgType
		S_StartSound(target, P_RandomRange(sfx_frpai1, sfx_frpai5))
        DoDamageAndBursts(target, nil, baseDamage, baseType)
        return true
    end

	if (HL.DoTDForestAccomodations and inf.type == MT_TAILSDOLL) then
		target.player.rings = $ + 1
	end

    -- damage from a valid source
    if not (src and src.valid) then
        return true
    end

    -- Figure out which mobj should count as the “inflictor”
    local stats   = HL1_DMGStats[src.type]
    local chosenInf = (stats and stats.damage and stats.damage.preferaggressor) and src or inf

    -- Play hurt sound
    S_StartSound(target, P_RandomRange(sfx_frpai1, sfx_frpai5))

    -- Award 50 points for non-self-inflicted, non-eggman-monitor kills
    if not (dmgType & DMG_DEATHMASK)
       and chosenInf
       and chosenInf.type ~= MT_EGGMAN_ICON
       and chosenInf.player
    then
        P_AddPlayerScore(chosenInf.player, 50)
    end

    -- If inf is nil after choosing, bail out (no further stats)
    if not chosenInf then
        return true
    end

    -- Determine damage and damage type
    local finalDamage = (not (chosenInf.ignoredamagedef) and HL_GetDamage(chosenInf)) or dmg
    local finalType   =
        (HL1_DMGStats[chosenInf.type] and HL1_DMGStats[chosenInf.type].damagetype)
        or damageTypeMaps[dmgType & ~DMG_CANHURTSELF]
        or dmgType

    -- Warn and seed a default entry if we have no stats for this aggressor
    if not HL1_DMGStats[chosenInf.type] then
        CONS_Printf(target.player, "No DMGStats found for aggressor!")
        HL1_DMGStats[chosenInf.type] = { damage = { dmg = 15 } }
    end

    -- Give invuln if this wasn't from another Freeman
    if not chosenInf.stats then
        target.player.powers[pw_flashing] = 18
    end
    target.player.timeshit = target.player.timeshit + 1

    -- Finally, do the damage and bursts
    DoDamageAndBursts(target, chosenInf, finalDamage, finalType)
    return true

end, MT_PLAYER)

-- Helper to extract the numeric suffix from a sentinel string, e.g. "CROWBAR_SWING_1" => 1
local function getSentinelNumber(s)
	local num = s:match("(%d+)$")
	return tonumber(num) or 0
end

local FULL_CIRCLE = ANGLE_MAX  >> 16
local HALF_CIRCLE = FULL_CIRCLE / 2

local function SignedAngle16(a)
    -- convert unsigned [0 … 65535] into signed [–32768 … +32767]
    if a >= HALF_CIRCLE then
        return a - FULL_CIRCLE
    else
        return a
    end
end

local function UnsignedAngle16(a)
    if a < 0 then
        return a + FULL_CIRCLE
    else
        return a
    end
end

addHook("PlayerThink", function(player)
    if not player.mo or player.mo.skin ~= skin then return end

    -- Decay HL1-style punch angle (viewkick)
    if player.hl.punchangle then
        local punch = player.hl.punchangle

        -- Turn vector into "true-value"
		local x = SignedAngle16(punch.x or 0)
		local y = SignedAngle16(punch.y or 0)
		local z = SignedAngle16(punch.z or 0)
        -- Compute magnitude len = sqrt(x^2 + y^2 + z^2)
        local len = FixedHypot(FixedHypot(x, y), z)
        if len > 0 then
            -- Normalize: nx,ny,nz are unit‐length in fixed
            local nx = FixedDiv(x, len)
            local ny = FixedDiv(y, len)
            local nz = FixedDiv(z, len)

            -- Decay factor: (10.0 + len*0.5) * frametime
            --   -> (10*FRACUNIT + (len/2)) * (FRACUNIT/TICRATE)
            local baseDecay   = ((10 * ANG1) >> 16)
            local variablePart = FixedDiv(len, 2*FRACUNIT)    -- len * 0.5
            local decayPerSec  = baseDecay + variablePart
            local frametimeFP  = FRACUNIT/TICRATE  -- = FRACUNIT*(1/35)
            local decay        = FixedMul(decayPerSec, frametimeFP)

            -- Subtract and clamp
            len = len - decay
            if len < 0 then len = 0 end

            -- Re‑scale punch vector
            punch.x = UnsignedAngle16(FixedMul(nx, len))
            punch.y = UnsignedAngle16(FixedMul(ny, len))
            punch.z = UnsignedAngle16(FixedMul(nz, len))
        end
    end

    -- Increment Damage Tics Clock
    player.hl1damagetics = (player.hl1damagetics or 0) + 1
end)

local function printTable(data, prefix)
	prefix = prefix or ""
	if type(data) == "table"
		for k, v in pairs(data or {}) do
			local key = prefix .. k
			if type(v) == "table" then
				print("key " .. key .. " = a table:")
				printTable(v, key .. ".")
			else
				print("key " .. key .. " = " .. v)
			end
		end
	else
		print(data)
	end
end

addHook("PlayerThink", function(player)
	if not player.mo or player.mo.skin ~= skin then return end

	local anim = player.hl1currentAnimation
	if not anim then
		HL_ChangeViewmodelState(player, "idle", "idle")
		anim = player.hl1currentAnimation
		if not anim then return end
	end

	-- tick down
	if player.hl1frameclock > 0 then
		player.hl1frameclock = player.hl1frameclock - 1
		return
	end

	-- advance frame
	player.hl1frame = player.hl1frame + 1

	-- pick up any sound
	local s = anim.frameSounds and anim.frameSounds[player.hl1frame]
	if s then
		if type(s)=="table" then
			S_StartSound(player.mo, s.sound + P_RandomRange(0, s.sounds-1))
		else
			S_StartSound(player.mo, s)
		end
	end

	-- find the highest frame index in frameDurations
	local maxFrame = 0
	for idx in pairs(anim.frameDurations) do
		maxFrame = max(maxFrame, idx)
	end

	if player.hl1frame > maxFrame then
		-- we've run out of frames
		if anim.loop then
			-- wrap back to frame 1
			player.hl1frame = 1
			player.hl1frameclock = anim.frameDurations[1] or 1
			return
		elseif anim.next then
			-- jump into the next state
			return HL_ChangeViewmodelState(player, anim.next, anim.next)
		else
			-- no loop, no next: fall back to idle
			return HL_ChangeViewmodelState(player, "idle", "idle")
		end
	end

	-- still in bounds: set clock for this new frame
	-- find the duration for the highest index <= current frame
	local dur = 1
	for idx, d in ipairs(anim.frameDurations) do
		if idx <= player.hl1frame then dur = d else break end
	end
	player.hl1frameclock = dur
end)

local function contains(mainStr, subStr)
	return string.find(mainStr, subStr) ~= nil
end

addHook("PlayerThink", function(player)
	if not player.mo return end
	if player.mo.skin != skin return end
	local viewmodel = kombihl1viewmodels[HL_WpnStats[player.hl.curwep].viewmodel or "PISTOL"]
	if not player.hlinv.wepclips[player.hl.curwep]
		player.hlinv.wepclips[player.hl.curwep] = {
			HL_WpnStats[player.hl.curwep].primary and HL_WpnStats[player.hl.curwep].primary.clipsize or -1,
			HL_WpnStats[player.hl.curwep].secondary and HL_WpnStats[player.hl.curwep].secondary.clipsize or -1
		}
	end

	-- Set-up reloading
	local weapon_stats = HL_WpnStats[player.hl.curwep]
	local weapon_clips = player.hlinv.wepclips[player.hl.curwep]
	local primary = weapon_stats.primary or HL_WpnStats["9mmhandgun"].primary
	local ammo_type = primary.ammo
	local reload_increment = primary.reloadincrement

	if ((not weapon_clips.primary
		or (((player.cmd.buttons & BT_CUSTOM1) or (player.hlcmds and player.hlcmds.reload)) and weapon_clips.primary < primary.clipsize))) and contains(player.hl1viewmdaction, "idle")
		and not player.kombireloading
		and player.hl1weapondelay == 0
		and (player.hlinv.ammo[ammo_type] or 0) > 0

		if primary.reloadsound then
			S_StartSound(player.mo, primary.reloadsound)
		end
		player.kombireloading = 1
		player.hl1weapondelay = weapon_stats.globalfiredelay.reloadstart or weapon_stats.globalfiredelay.reload
		HL_ChangeViewmodelState(player, "reload start", "idle 1")
	end

	-- Now do the reloading
	if player.kombireloading == 1 and player.hl1weapondelay == 0
		if primary.clipsize and ammo_type
			if not weapon_clips.primary weapon_clips.primary = 0 end
			local max_reload = primary.clipsize - weapon_clips.primary
			local available_ammo = player.hlinv.ammo[ammo_type]

			if reload_increment
				if player.hl1doreload
					local to_reload = min(reload_increment, max_reload, available_ammo)
					weapon_clips.primary = $ + to_reload
					player.hlinv.ammo[ammo_type] = $ - to_reload
				end

				if weapon_clips.primary >= weapon_stats.primary.clipsize or player.hlinv.ammo[ammo_type] <= 0
					HL_ChangeViewmodelState(player, "reload stop", "idle 1")
					player.kombireloading = 0 -- Stop reloading if full or out of ammo
					player.hl1doreload = nil -- de-init this variable.
				else
					HL_ChangeViewmodelState(player, "reload loop", "idle 1")
					player.hl1weapondelay = weapon_stats.globalfiredelay.reloadloop or weapon_stats.globalfiredelay.reload -- Continue reloading incrementally
					player.hl1doreload = true -- allow reloading to start
				end
			else
				local wasempty = weapon_clips.primary <= 0
				local to_reload = min(max_reload, available_ammo)
				player.hlinv.ammo[ammo_type] = $ - to_reload
				weapon_clips.primary = $ + to_reload
				local wepstats = weapon_stats or HL_WpnStats["9mmhandgun"]
				player.kombireloading = 0
				player.hl1weapondelay = wasempty and wepstats.globalfiredelay.reloadpost or 0
			end
		else
			print("Weapon " .. player.hl.curwep .. " missing necessary stats! Check 'clipsize' and 'ammo'.")
		end
	end
end)