local function SafeFreeSlot(...)
	for _,slot in ipairs({...})
		if not rawget(_G, slot) freeslot(slot) end -- overlapping = wasting, how do we not waste (as many of) them? don't do it in the first place!
	end
end

local gsrcobjects = {
  [0] = {
    name  = "item_suit",
    gifts = {
      -- default gift: you always give “suit = true”
      suit = true,
    },
    flag_gifts = {
      -- if bit 1 (0x1) is set in arg1, then merge in `{isshort = true}`
      [0x1] = { key = "suit", params = { shortlogon = true } },
    }
  },
  [1] = {
    name  = "item_longjump",
    gifts = {
      longjump = true,
    },
  },
}

SafeFreeSlot("S_GSRCOBJECT", "MT_GSRCOBJECTSPAWN")

states[S_GSRCOBJECT] = {
	sprite = SPR_PLAY,
	frame = A,
	tics = -1,
	nextstate = S_GSRCOBJECT,
}

mobjinfo[MT_GSRCOBJECTSPAWN] = {
--$Title GoldSrc Object Spawner
--$Sprite PLAYA0
--$Category Half-Life
--$ParameterText Object Type
--$Arg0 "Object Type"
--$Arg0Type 11
--$Arg0Enum {0="item_suit"; 1="item_longjump"; }
--$Arg0Default 0
--$Arg0ToolTip "Sets what item will spawn."
--$Angled
doomednum = 2051,
spawnstate = S_GSRCOBJECT,
spawnhealth = 1000,
deathstate = S_GSRCOBJECT,
radius = 6*FRACUNIT,
height = 12*FRACUNIT,
dispoffset = 3,
flags = MF_SCENERY|MF_SPECIAL,
}
-- Max seems to be 36?

-- 2) Hook into spawn, read both args[0] (object type) and args[1] (bitflags),
--    then build a fresh gifts table by copying the defaults and merging in
--    anything whose bit is set.
addHook("MapThingSpawn", function(mobj, mapthing)
    local def      = gsrcobjects[mapthing.args[0]]
    local bitmask  = mapthing.args[1] or 0
    local gifts    = {}

    -- copy the base gifts in (so gifts.suit = true, etc)
    for k,v in pairs(def.gifts) do
      gifts[k] = v
    end

    -- now for each defined flag‐entry, if the bit is on, merge it
    for bit,info in pairs(def.flag_gifts or {}) do
      if bitmask & bit ~= 0 then        -- Lua 5.3 bitwise test
        -- if the key already exists, overwrite or augment:
        if type(gifts[info.key]) == "table" then
          -- merge fields
          for pk,pv in pairs(info.params) do
            gifts[info.key][pk] = pv
          end
        else
          -- brand-new table entry
          gifts[info.key] = info.params
        end
      end
    end

    -- finally assign it
    mobj.pickupstats = gifts
end, MT_GSRCOBJECTSPAWN)